/*
Copyright (c) 2005 Health Market Science, Inc.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
USA

You can contact Health Market Science at info@healthmarketscience.com
or at the following address:

Health Market Science
2700 Horizon Drive
Suite 200
King of Prussia, PA 19406
*/

package com.healthmarketscience.jackcess;

import java.sql.SQLException;
import java.sql.Types;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;

/**
 * Access data type
 * @author Tim McCune
 */
public class DataType {
  
  public static final DataType BOOLEAN = new DataType((byte) 0x01, Types.BOOLEAN, 0);
  public static final DataType BYTE = new DataType((byte) 0x02, Types.TINYINT, 1);
  public static final DataType INT = new DataType((byte) 0x03, Types.SMALLINT, 2);
  public static final DataType LONG = new DataType((byte) 0x04, Types.INTEGER, 4);
  public static final DataType MONEY = new DataType((byte) 0x05, Types.DECIMAL, 8);
  public static final DataType FLOAT = new DataType((byte) 0x06, Types.FLOAT, 4);
  public static final DataType DOUBLE = new DataType((byte) 0x07, Types.DOUBLE, 8);
  public static final DataType SHORT_DATE_TIME = new DataType((byte) 0x08, Types.TIMESTAMP, 8);
  public static final DataType BINARY = new DataType((byte) 0x09, Types.BINARY, 255, true);
  public static final DataType TEXT = new DataType((byte) 0x0A, Types.VARCHAR, 50 * 2, true);
  public static final DataType OLE = new DataType((byte) 0x0B, Types.LONGVARBINARY, 12, true);
  public static final DataType MEMO = new DataType((byte) 0x0C, Types.LONGVARCHAR, 12, true);
  public static final DataType UNKNOWN_0D = new DataType((byte) 0x0D);
  //public static final DataType GUID = new DataType((byte) 0x0F, null, 16);
  public static final DataType GUID = new DataType((byte) 0x0F, 0, 16);
  public static final DataType NUMERIC = new DataType((byte) 0x10, Types.NUMERIC, 17);

  public static final Collection values = new HashSet();
  static {
	  values.add(BOOLEAN);
	  values.add(BYTE);
	  values.add(INT);
	  values.add(LONG);
	  values.add(MONEY);
	  values.add(FLOAT);
	  values.add(DOUBLE);
	  values.add(SHORT_DATE_TIME);
	  values.add(BINARY);
	  values.add(TEXT);
	  values.add(OLE);
	  values.add(MEMO);
	  values.add(UNKNOWN_0D);
	  values.add(GUID);
	  values.add(NUMERIC);
  }
  public static final Collection values()
  {
	  return values;
  }
  
  /** Map of SQL types to Access data types */
  private static Map SQL_TYPES = new HashMap();
  static {
    //for (DataType type : DataType.values()) {
	for (Iterator it = DataType.values().iterator(); it.hasNext();) {
	  DataType type = (DataType)it.next(); 
      if (type._sqlType != null) {
        SQL_TYPES.put(type._sqlType, type);
      }
    }
    SQL_TYPES.put(new Integer(Types.BIT), BYTE);
    SQL_TYPES.put(new Integer(Types.BLOB), OLE);
    SQL_TYPES.put(new Integer(Types.BIGINT), LONG);
    SQL_TYPES.put(new Integer(Types.CHAR), TEXT);
    SQL_TYPES.put(new Integer(Types.DATE), SHORT_DATE_TIME);
    SQL_TYPES.put(new Integer(Types.REAL), DOUBLE);
    SQL_TYPES.put(new Integer(Types.TIME), SHORT_DATE_TIME);
    SQL_TYPES.put(new Integer(Types.VARBINARY), BINARY);
  }
  
  private static Map DATA_TYPES = new HashMap();
  static {
     for (Iterator it = DataType.values().iterator(); it.hasNext();) {
  	  DataType type = (DataType)it.next(); 
        if (type._sqlType != null) {
        	DATA_TYPES.put(new Integer(type._value), type);
        }
      }
  }

  /** is this a variable length field */
  private boolean _variableLength;
  /** Internal Access value */
  private byte _value;
  /** Size in bytes */
  private Integer _size;
  /** SQL type equivalent, or null if none defined */
  private Integer _sqlType;
  
  private DataType(byte value) {
    this(value, null, null);
  }
  
  private DataType(byte value, int sqlType, int size) {
	    this(value, new Integer(sqlType), new Integer(size), false);
	  }
  
  private DataType(byte value, Integer sqlType, Integer size) {
    this(value, sqlType, size, false);
  }
  
  private DataType(byte value, Integer sqlType, Integer size,
                   boolean variableLength) {
    _value = value;
    _sqlType = sqlType;
    _size = size;
    _variableLength = variableLength;
  }
  private DataType(byte value, int sqlType, int size,
          boolean variableLength) {
	  this(value, new Integer(sqlType), new Integer(size), variableLength);
  }
  
  public byte getValue() {
    return _value;
  }
  
  public boolean isVariableLength() {
    return _variableLength;
  }
  
  public int getSize() {
    if (_size != null) {
      return _size.intValue();
    } else {
      throw new IllegalArgumentException("FIX ME");
    }
  }
  
  public int getSQLType() throws SQLException {
    if (_sqlType != null) {
      return _sqlType.intValue();
    } else {
      throw new SQLException("Unsupported data type: " + toString());
    }
  }
  
  public static DataType fromByte(byte b) throws SQLException {
    DataType rtn = (DataType) DATA_TYPES.get(new Integer(b));
    if (rtn != null) {
      return rtn;
    } else {
      throw new SQLException("Unrecognized data type: " + b);
    }
  }
  
  public static DataType fromSQLType(int sqlType) throws SQLException {
    DataType rtn = (DataType) SQL_TYPES.get(new Integer(sqlType));
    if (rtn != null) {
      return rtn;
    } else {
      throw new SQLException("Unsupported SQL type: " + sqlType);
    }
  }
  
}
